package org.iGEM.TU_Eindhoven;

import java.awt.GridLayout;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;

import org.apache.commons.math3.special.Erf;
import org.iGEM.TU_Eindhoven.Particle.particleType;


public class Main {
	static Random rand = new Random();
	static double fieldSize;
	static double Rm;
	static double forsterDist;
	static double maxForsterDist;
	static List<Particle> particleList= new ArrayList<Particle>();
	static int particleAmount;
	static int steps;

	static double deviationxy1;
	static double deviationz1;
	static double meanZ1;
	static double deviationxy2;
	static double deviationz2;
	static double meanZ2;

	private static PrintWriter out;

	static Thread timer;
	static int currentStep;

	public static void main(String[] args) throws IOException {	
	
		JPanel panel = new JPanel(new GridLayout(0, 1));
		panel.add(new JLabel("Note: multiple amounts of particles can be simulated at once using a comma."));
		panel.add(new JLabel(""));
		JTextField field1 = new JTextField("");
		JTextField field2 = new JTextField("");
		JTextField field3 = new JTextField("");
		JTextField field4 = new JTextField("");
		JTextField field5 = new JTextField("");
		JTextField field6 = new JTextField("");
		JTextField field7 = new JTextField("");
		JTextField field8 = new JTextField("");
		JTextField field9 = new JTextField("");
		JTextField field10 = new JTextField("");
		JTextField field11 = new JTextField("");
		

		panel.add(new JLabel("Box Size (nm):"));
		panel.add(field1);
		panel.add(new JLabel("Amount of Particles:"));
		panel.add(field2);
		panel.add(new JLabel("Radius of Membrane Protein (nm):"));
		panel.add(field3);
		panel.add(new JLabel("Length of Linker (nm):"));
		panel.add(field4);
		panel.add(new JLabel("Radius of First Tethered Particle (nm)"));
		panel.add(field5);
		panel.add(new JLabel("Radius of Second Tethered Particle (nm)"));
		panel.add(field6);
		panel.add(new JLabel("Persistence Length of Linker (nm)"));
		panel.add(field7);
		panel.add(new JLabel("Förster Distance (nm):"));
		panel.add(field8);
		panel.add(new JLabel("Max Förster Distance (nm):"));
		panel.add(field9);
		panel.add(new JLabel("Iterations:"));
		panel.add(field10);
		panel.add(new JLabel("Output Field:"));
		panel.add(field11);
	
		int result = JOptionPane.showConfirmDialog(null, panel, "Please fill in the parameters",
				JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);
		if(result != 0){
			System.exit(0);
		}

		fieldSize = Double.valueOf(field1.getText());	
		String[] amounts = field2.getText().replaceAll(" ", "").split(",");
		Rm = Double.valueOf(field3.getText());
		double L = Double.valueOf(field4.getText());
		double Rp1 = Double.valueOf(field5.getText());
		double Rp2 = Double.valueOf(field6.getText());
		double P = Double.valueOf(field7.getText());
		forsterDist = Double.valueOf(field8.getText());
		maxForsterDist = Double.valueOf(field9.getText());
		steps = Integer.valueOf(field10.getText());
		out = new PrintWriter(field11.getText(),"UTF-8");


		System.out.println("Box size: " + fieldSize + "nm");
		System.out.println("Radius of Membrane Protein: " + Rm + "nm");
		System.out.println("Length of Linker: " + L + "nm");
		System.out.println("Radius of First Tethered Particle: " + Rp1 + "nm");
		System.out.println("Radius of Second Tethered Particle: " + Rp2 + "nm");
		System.out.println("Persistence Length of Linker " + P + "nm");
		System.out.println("Förster Distance " + forsterDist + "nm");
		System.out.println("Max Förster Distance " + maxForsterDist + "nm");
		System.out.println("Iterations " + steps);

		double Nr = Rp1/Math.sqrt(P*L/3);                    //calculate linker properties for both linkers, like mean values and deviation
		double erfNr = Erf.erf(Nr);
		deviationxy1 = Math.sqrt((P*L/3)*(2+((4*Nr)/(Math.sqrt(Math.PI) * erfNr))));
		deviationz1 = Math.sqrt((P*L/3)*(2+((4*Nr)/(Math.sqrt(Math.PI) * erfNr)) + Math.pow(Nr, 2)));
		meanZ1 = Math.sqrt(L*P/3)*(((2*(1- Math.pow(Math.E, -(Math.pow(Nr, 2)))))/(Math.sqrt(Math.PI)*erfNr)) + Nr * ((2-erfNr)/(erfNr)));

		Nr = Rp2/Math.sqrt(P*L/3);
		erfNr = Erf.erf(Nr);
		deviationxy2 = Math.sqrt((P*L/3)*(2+((4*Nr)/(Math.sqrt(Math.PI) * erfNr))));
		deviationz2 = Math.sqrt((P*L/3)*(2+((4*Nr)/(Math.sqrt(Math.PI) * erfNr)) + Math.pow(Nr, 2)));
		meanZ2 = Math.sqrt(L*P/3)*(((2*(1- Math.pow(Math.E, -(Math.pow(Nr, 2)))))/(Math.sqrt(Math.PI)*erfNr)) + Nr * ((2-erfNr)/(erfNr)));


		for(String amount : amounts){
			System.out.println("Starting Simulation with "+ amount + " Particles");
			particleList.clear();
			particleAmount = Integer.valueOf(amount);
			out.print(particleAmount + ",");
			out.flush();
			initiate();			//The particles are placed
			simulate();			//The linker is simulated and BRET or FRET is calculated
		}
		out.close();
	}

	//Method to print the progress every 2 seconds
	private static void planSched(){
		timer = new Thread(new Runnable() {
			public void run() {
				while(true){
					try {
						Thread.sleep(2000);
					} catch (InterruptedException e) {
					}
					System.out.println(Main.currentStep/(float)steps*100 + "%");	
				}
			}
		});
		timer.start();

	}

	@SuppressWarnings("deprecation")
	private static void simulate() {
		double totalBRET = 0; 		
		currentStep = 1;  
		planSched();
		for(currentStep = 1; currentStep <= steps ; currentStep++){
			//Check for interaction
			List<Particle> interactionCheck = new ArrayList<Particle>();
			interactionCheck.addAll(particleList);				//All particles must be checked at the start
			boolean changed = true;								//Changed is true if existing BRET pairs are broken. Now it is true to enter the loop
			while(changed){
				changed = false;
				int size = interactionCheck.size();
				for(int j = 0; j < size; j++){
					Particle A = interactionCheck.get(j);
					for(int k = j+1; k < size; k++){
						Particle B = interactionCheck.get(k);
						if (A.getType() == B.getType()){continue;}  //If both are the same type, two BRET/FRET donors or two BRET/FRET acceptors, skip the check.
						double distance = distance(A.getX(),B.getX(),A.getY(),B.getY());
						if (distance <= (maxForsterDist)){			//Only check for a new pair if the distance is less than the max forster distance.
							if(A.BRET == false && B.BRET == false){ //If both particles don't form a pair already with an other particle, create a new pair.
								A.distance = distance;
								B.distance = distance;

								A.BRETPartner = B;
								B.BRETPartner = A;

								A.BRET = true;
								B.BRET = true;
							}
							else if(A.BRET == true && B.BRET == false && distance < A.distance){	//If A already forms a pair, but B is closer, form a pair with B and set changed = true. 
								changed = true;														//Because A already formed a partner, changed must be set to true. In the next iteration this former partner can look for a new BRET pair

								A.BRETPartner.BRET = false;

								A.distance = distance;
								B.distance = distance;

								A.BRETPartner = B;
								B.BRETPartner = A;

								A.BRET = true;
								B.BRET = true;
							}
							else if(A.BRET == false && B.BRET == true && distance < B.distance){ //Same for particle B
								changed = true;

								B.BRETPartner.BRET = false;

								A.distance = distance;
								B.distance = distance;

								A.BRETPartner = B;
								B.BRETPartner = A;

								A.BRET = true;
								B.BRET = true;
							}
							else if(A.BRET == true && B.BRET == true && distance < A.distance && distance < B.distance){ //If both already had formed a pair, but this is a better pair
								changed = true;

								A.BRETPartner.BRET = false;
								B.BRETPartner.BRET = false;

								A.distance = distance;
								B.distance = distance;

								A.BRETPartner = B;
								B.BRETPartner = A;

								A.BRET = true;
								B.BRET = true;
							}


						}
					}
				}
				for(int j = size-1; j >= 0; j--){				//At the end of each iteration, all particles that form a pair, form the best pair. These can be removed from the checklist.
					if(interactionCheck.get(j).BRET == true){
						interactionCheck.remove(j);
					}
				}
			}

			//Calculate BRET

			for(Particle A : particleList){			//Check for every particle the BRET or FRET contribution
				boolean B = A.BRET;
				A.BRET = false;
				if(A.getType() == particleType.B || B == false){continue;} //Only check one particle per pair

				double dist = A.distance;

				//linker simulation

				double angle = rand.nextDouble()*2*Math.PI;

				double projectionRadius1 = rand.nextGaussian() * deviationxy1; // calculate the radius in the XY plane (lateral to the membrane)
				double projectionRadiusX1 = Math.sin(angle) * projectionRadius1; //calculate the translation in the X-axis
				double projectionRadiusY1 = Math.cos(angle) * projectionRadius1; //same for the Y-axis
				double zRadius1 = (rand.nextGaussian() * deviationz1) + meanZ1; //Calculate Z position (depth)

				angle = rand.nextDouble()*2*Math.PI;

				double projectionRadius2 = rand.nextGaussian() * deviationxy2;//Do it again for the second membrane protein
				double projectionRadiusX2 = Math.sin(angle) * projectionRadius2 + dist; //Add the distance between the membrane proteins
				double projectionRadiusY2 = Math.cos(angle) * projectionRadius2;
				double zRadius2 = (rand.nextGaussian() * deviationz2) + meanZ2;
				//Calculate the distance between the two intracellular parts
				dist = Math.sqrt(Math.pow(projectionRadiusX1-projectionRadiusX2, 2) + Math.pow(projectionRadiusY1-projectionRadiusY2, 2) + Math.pow(zRadius1-zRadius2, 2));

				//Calculate BRET or FRET and add it to the total
				totalBRET += 1d/(1d+Math.pow(dist/forsterDist, 6));

			}
		}
		timer.stop();
		System.out.println("Done!");

		//Print the mean BRET or FRET
		out.println(totalBRET / (particleAmount/2d*steps));
		out.flush();
	}
	
	//Method to calculate distance between two particles
	static double distance(double X1, double X2, double Y1, double Y2){ 
		double xDiff = Math.abs(X1-X2);
		if(xDiff > fieldSize/2){xDiff = fieldSize - xDiff;}
		double yDiff = Math.abs(Y1-Y2);
		if(yDiff > fieldSize/2){yDiff = fieldSize - yDiff;}
		double distance = Math.sqrt(Math.pow(xDiff, 2) + Math.pow(yDiff, 2));
		return distance;
	}

	//
	private static void initiate() {
		//Create particles in equal amounts of the two variants
		for(int i = 0; i < particleAmount; i++){
			if(i%2 == 0){
				new Particle(Particle.particleType.A);
			}
			else{
				new Particle(Particle.particleType.B);
			}
		}
	}
}
